@@ -14,21 +14,22 @@
 # Copyright Buildbot Team Members
 
 import collections
+import json
 import re
-import warnings
 import weakref
 
+from twisted.internet import defer
+from twisted.python.components import registerAdapter
+from zope.interface import implementer
+
 from buildbot import config
 from buildbot import util
 from buildbot.interfaces import IProperties
 from buildbot.interfaces import IRenderable
 from buildbot.util import flatten
-from buildbot.util import json
-from twisted.internet import defer
-from twisted.python.components import registerAdapter
-from zope.interface import implements
 
 
+@implementer(IProperties)
 class Properties(util.ComparableMixin):
 
     """
@@ -47,7 +48,6 @@ class Properties(util.ComparableMixin):
     """
 
     compare_attrs = ('properties',)
-    implements(IProperties)
 
     def __init__(self, **kwargs):
         """
@@ -58,13 +58,67 @@ class Properties(util.ComparableMixin):
         # persisted if a build is rebuilt
         self.runtime = set()
         self.build = None  # will be set by the Build when starting
+        self._used_secrets = {}
         if kwargs:
             self.update(kwargs, "TEST")
+        self._master = None
+        self._sourcestamps = None
+        self._changes = None
+
+    @property
+    def master(self):
+        if self.build is not None:
+            return self.build.master
+        return self._master
+
+    @master.setter
+    def master(self, value):
+        self._master = value
+
+    @property
+    def sourcestamps(self):
+        if self.build is not None:
+            return [b.asDict() for b in self.build.getAllSourceStamps()]
+        elif self._sourcestamps is not None:
+            return self._sourcestamps
+        raise AttributeError('neither build nor _sourcestamps are set')
+
+    @sourcestamps.setter
+    def sourcestamps(self, value):
+        self._sourcestamps = value
+
+    def getSourceStamp(self, codebase=''):
+        for source in self.sourcestamps:
+            if source['codebase'] == codebase:
+                return source
+        return None
+
+    @property
+    def changes(self):
+        if self.build is not None:
+            return [c.asChDict() for c in self.build.allChanges()]
+        elif self._changes is not None:
+            return self._changes
+        raise AttributeError('neither build nor _changes are set')
+
+    @changes.setter
+    def changes(self, value):
+        self._changes = value
+
+    @property
+    def files(self):
+        if self.build is not None:
+            return self.build.allFiles()
+        files = []
+        # self.changes, not self._changes to raise AttributeError if unset
+        for chdict in self.changes:
+            files.extend(chdict['files'])
+        return files
 
     @classmethod
     def fromDict(cls, propDict):
         properties = cls()
-        for name, (value, source) in propDict.iteritems():
+        for name, (value, source) in propDict.items():
             properties.setProperty(name, value, source)
         return properties
 
@@ -86,24 +140,25 @@ class Properties(util.ComparableMixin):
         rv = self.properties[name][0]
         return rv
 
-    def __nonzero__(self):
-        return not not self.properties
+    def __bool__(self):
+        return bool(self.properties)
 
     def getPropertySource(self, name):
         return self.properties[name][1]
 
     def asList(self):
         """Return the properties as a sorted list of (name, value, source)"""
-        l = sorted([(k, v[0], v[1]) for k, v in self.properties.iteritems()])
-        return l
+        ret = sorted([(k, v[0], v[1]) for k, v in self.properties.items()])
+        return ret
 
     def asDict(self):
-        """Return the properties as a simple key:value dictionary"""
-        return dict(self.properties)
+        """Return the properties as a simple key:value dictionary,
+        properly unicoded"""
+        return self.properties.copy()
 
     def __repr__(self):
         return ('Properties(**' +
-                repr(dict((k, v[0]) for k, v in self.properties.iteritems())) +
+                repr(dict((k, v[0]) for k, v in self.properties.items())) +
                 ')')
 
     def update(self, dict, source, runtime=False):
@@ -119,7 +174,7 @@ class Properties(util.ComparableMixin):
     def updateFromPropertiesNoRuntime(self, other):
         """Update this object based on another object, but don't
         include properties that were marked as runtime."""
-        for k, v in other.properties.iteritems():
+        for k, v in other.properties.items():
             if k not in other.runtime:
                 self.properties[k] = v
 
@@ -134,13 +189,10 @@ class Properties(util.ComparableMixin):
     has_key = hasProperty
 
     def setProperty(self, name, value, source, runtime=False):
-        try:
-            json.dumps(value)
-        except TypeError:
-            warnings.warn(
-                "Non jsonable properties are not explicitly supported and" +
-                "will be explicitly disallowed in a future version.",
-                DeprecationWarning, stacklevel=2)
+        name = util.bytes2unicode(name)
+        if not IRenderable.providedBy(value):
+            json.dumps(value)  # Let the exception propagate ...
+        source = util.bytes2unicode(source)
 
         self.properties[name] = (value, source)
         if runtime:
@@ -156,12 +208,28 @@ class Properties(util.ComparableMixin):
         renderable = IRenderable(value)
         return defer.maybeDeferred(renderable.getRenderingFor, self)
 
+    # as the secrets are used in the renderable, they can pretty much arrive anywhere
+    # in the log of state strings
+    # so we have the renderable record here which secrets are used that we must remove
+    def useSecret(self, secret_value, secret_name):
+        if secret_value.strip():
+            self._used_secrets[secret_value] = "<" + secret_name + ">"
+
+    # This method shall then be called to remove secrets from any text that could be logged
+    # somewhere and that could contain secrets
+    def cleanupTextFromSecrets(self, text):
+        # Better be correct and inefficient than efficient and wrong
+        secrets = self._used_secrets
+        for k in sorted(secrets, key=len, reverse=True):
+            text = text.replace(k, secrets[k])
+        return text
+
 
 class PropertiesMixin:
 
     """
     A mixin to add L{IProperties} methods to a class which does not implement
-    the interface, but which can be coerced to the interface via an adapter.
+    the full interface, only getProperties() function.
 
     This is useful because L{IProperties} methods are often called on L{Build}
     and L{BuildStatus} objects without first coercing them.
@@ -173,32 +241,98 @@ class PropertiesMixin:
     set_runtime_properties = False
 
     def getProperty(self, propname, default=None):
-        props = IProperties(self)
-        return props.getProperty(propname, default)
+        return self.getProperties().getProperty(propname, default)
 
     def hasProperty(self, propname):
-        props = IProperties(self)
-        return props.hasProperty(propname)
+        return self.getProperties().hasProperty(propname)
 
     has_key = hasProperty
 
     def setProperty(self, propname, value, source='Unknown', runtime=None):
         # source is not optional in IProperties, but is optional here to avoid
         # breaking user-supplied code that fails to specify a source
-        props = IProperties(self)
+        props = self.getProperties()
         if runtime is None:
             runtime = self.set_runtime_properties
         props.setProperty(propname, value, source, runtime=runtime)
 
-    def getProperties(self):
-        return IProperties(self)
-
     def render(self, value):
-        props = IProperties(self)
-        return props.render(value)
+        return self.getProperties().render(value)
 
 
-class _PropertyMap(object):
+@implementer(IRenderable)
+class RenderableOperatorsMixin:
+
+    """
+    Properties and Interpolate instances can be manipulated with standard operators.
+    """
+
+    def __eq__(self, other):
+        return _OperatorRenderer(self, other, "==", lambda v1, v2: v1 == v2)
+
+    def __ne__(self, other):
+        return _OperatorRenderer(self, other, "!=", lambda v1, v2: v1 != v2)
+
+    def __lt__(self, other):
+        return _OperatorRenderer(self, other, "<", lambda v1, v2: v1 < v2)
+
+    def __le__(self, other):
+        return _OperatorRenderer(self, other, "<=", lambda v1, v2: v1 <= v2)
+
+    def __gt__(self, other):
+        return _OperatorRenderer(self, other, ">", lambda v1, v2: v1 > v2)
+
+    def __ge__(self, other):
+        return _OperatorRenderer(self, other, ">=", lambda v1, v2: v1 >= v2)
+
+    def __add__(self, other):
+        return _OperatorRenderer(self, other, "+", lambda v1, v2: v1 + v2)
+
+    def __sub__(self, other):
+        return _OperatorRenderer(self, other, "-", lambda v1, v2: v1 - v2)
+
+    def __mul__(self, other):
+        return _OperatorRenderer(self, other, "*", lambda v1, v2: v1 * v2)
+
+    def __truediv__(self, other):
+        return _OperatorRenderer(self, other, "/", lambda v1, v2: v1 / v2)
+
+    def __floordiv__(self, other):
+        return _OperatorRenderer(self, other, "//", lambda v1, v2: v1 // v2)
+
+    def __mod__(self, other):
+        return _OperatorRenderer(self, other, "%", lambda v1, v2: v1 % v2)
+
+    # we cannot use this trick to overload the 'in' operator, as python will force the result
+    # of __contains__ to a boolean, forcing it to True all the time
+    # so we mimic sqlalchemy and make a in_ method
+    def in_(self, other):
+        return _OperatorRenderer(self, other, "in", lambda v1, v2: v1 in v2)
+
+
+@implementer(IRenderable)
+class _OperatorRenderer(RenderableOperatorsMixin, util.ComparableMixin):
+    """
+    An instance of this class renders a comparison given by a operator
+    function with v1 and v2
+
+    """
+
+    compare_attrs = ('fn',)
+
+    def __init__(self, v1, v2, cstr, comparator):
+        self.v1, self.v2, self.comparator, self.cstr = v1, v2, comparator, cstr
+
+    @defer.inlineCallbacks
+    def getRenderingFor(self, props):
+        v1, v2 = yield props.render((self.v1, self.v2))
+        return self.comparator(v1, v2)
+
+    def __repr__(self):
+        return '%r %s %r' % (self.v1, self.cstr, self.v2)
+
+
+class _PropertyMap:
 
     """
     Privately-used mapping object to implement WithProperties' substitutions,
@@ -225,19 +359,18 @@ class _PropertyMap(object):
                 return self.temp_vals[prop]
             elif prop in properties:
                 return properties[prop]
-            else:
-                return repl
+            return repl
 
         def colon_tilde(mo):
             # %(prop:~repl)s
-            # if prop exists and is true (nonempty), use it; otherwise, use repl
+            # if prop exists and is true (nonempty), use it; otherwise, use
+            # repl
             prop, repl = mo.group(1, 2)
             if prop in self.temp_vals and self.temp_vals[prop]:
                 return self.temp_vals[prop]
             elif prop in properties and properties[prop]:
                 return properties[prop]
-            else:
-                return repl
+            return repl
 
         def colon_plus(mo):
             # %(prop:+repl)s
@@ -245,8 +378,7 @@ class _PropertyMap(object):
             prop, repl = mo.group(1, 2)
             if prop in properties or prop in self.temp_vals:
                 return repl
-            else:
-                return ''
+            return ''
 
         for regexp, fn in [
             (self.colon_minus_re, colon_minus),
@@ -275,6 +407,7 @@ class _PropertyMap(object):
         self.temp_vals[key] = val
 
 
+@implementer(IRenderable)
 class WithProperties(util.ComparableMixin):
 
     """
@@ -282,7 +415,6 @@ class WithProperties(util.ComparableMixin):
     want to interpolate build properties.
     """
 
-    implements(IRenderable)
     compare_attrs = ('fmtstring', 'args', 'lambda_subs')
 
     def __init__(self, fmtstring, *args, **lambda_subs):
@@ -290,11 +422,13 @@ class WithProperties(util.ComparableMixin):
         self.args = args
         if not self.args:
             self.lambda_subs = lambda_subs
-            for key, val in self.lambda_subs.iteritems():
+            for key, val in self.lambda_subs.items():
                 if not callable(val):
-                    raise ValueError('Value for lambda substitution "%s" must be callable.' % key)
+                    raise ValueError(
+                        'Value for lambda substitution "{}" must be callable.'.format(key))
         elif lambda_subs:
-            raise ValueError('WithProperties takes either positional or keyword substitutions, not both.')
+            raise ValueError(
+                'WithProperties takes either positional or keyword substitutions, not both.')
 
     def getRenderingFor(self, build):
         pmap = _PropertyMap(build.getProperties())
@@ -304,19 +438,32 @@ class WithProperties(util.ComparableMixin):
                 strings.append(pmap[name])
             s = self.fmtstring % tuple(strings)
         else:
-            for k, v in self.lambda_subs.iteritems():
+            for k, v in self.lambda_subs.items():
                 pmap.add_temporary_value(k, v(build))
             s = self.fmtstring % pmap
         return s
 
 
-_notHasKey = object()  # Marker object for _Lookup(..., hasKey=...) default
+class _NotHasKey(util.ComparableMixin):
+
+    """A marker for missing ``hasKey`` parameter.
+
+    To withstand ``deepcopy``, ``reload`` and pickle serialization round trips,
+    check it with ``==`` or ``!=``.
+    """
+    compare_attrs = ()
 
 
-class _Lookup(util.ComparableMixin, object):
-    implements(IRenderable)
+# any instance of _NotHasKey would do, yet we don't want to create and delete
+# them all the time
+_notHasKey = _NotHasKey()
 
-    compare_attrs = ('value', 'index', 'default', 'defaultWhenFalse', 'hasKey', 'elideNoneAs')
+
+@implementer(IRenderable)
+class _Lookup(util.ComparableMixin):
+
+    compare_attrs = (
+        'value', 'index', 'default', 'defaultWhenFalse', 'hasKey', 'elideNoneAs')
 
     def __init__(self, value, index, default=None,
                  defaultWhenFalse=True, hasKey=_notHasKey,
@@ -329,82 +476,120 @@ class _Lookup(util.ComparableMixin, object):
         self.elideNoneAs = elideNoneAs
 
     def __repr__(self):
-        return '_Lookup(%r, %r%s%s%s%s)' % (
-            self.value,
-            self.index,
-            ', default=%r' % (self.default,)
-            if self.default is not None else '',
-            ', defaultWhenFalse=False'
-            if not self.defaultWhenFalse else '',
-            ', hasKey=%r' % (self.hasKey,)
-            if self.hasKey is not _notHasKey else '',
-            ', elideNoneAs=%r' % (self.elideNoneAs,)
-            if self.elideNoneAs is not None else '')
+        return '_Lookup({}, {}{}{}{}{})'.format(
+            repr(self.value),
+            repr(self.index),
+            ', default={}'.format(repr(self.default)) if self.default is not None else '',
+            ', defaultWhenFalse=False' if not self.defaultWhenFalse else '',
+            ', hasKey={}'.format(repr(self.hasKey)) if self.hasKey != _notHasKey else '',
+            ', elideNoneAs={}'.format(repr(self.elideNoneAs))
+            if self.elideNoneAs is not None else ''
+            )
 
     @defer.inlineCallbacks
     def getRenderingFor(self, build):
         value = build.render(self.value)
         index = build.render(self.index)
         value, index = yield defer.gatherResults([value, index])
-        if not index in value:
+        if index not in value:
             rv = yield build.render(self.default)
         else:
             if self.defaultWhenFalse:
                 rv = yield build.render(value[index])
                 if not rv:
                     rv = yield build.render(self.default)
-                elif self.hasKey is not _notHasKey:
+                elif self.hasKey != _notHasKey:
                     rv = yield build.render(self.hasKey)
-            elif self.hasKey is not _notHasKey:
+            elif self.hasKey != _notHasKey:
                 rv = yield build.render(self.hasKey)
             else:
                 rv = yield build.render(value[index])
         if rv is None:
             rv = yield build.render(self.elideNoneAs)
-        defer.returnValue(rv)
+        return rv
 
 
 def _getInterpolationList(fmtstring):
     # TODO: Verify that no positional substitutions are requested
     dd = collections.defaultdict(str)
     fmtstring % dd
-    return dd.keys()
+    return list(dd)
 
 
-class _PropertyDict(object):
-    implements(IRenderable)
+@implementer(IRenderable)
+class _PropertyDict:
 
     def getRenderingFor(self, build):
         return build.getProperties()
+
+
 _thePropertyDict = _PropertyDict()
 
 
-class _SourceStampDict(util.ComparableMixin, object):
-    implements(IRenderable)
+@implementer(IRenderable)
+class _WorkerPropertyDict:
+
+    def getRenderingFor(self, build):
+        return build.getBuild().getWorkerInfo()
+
+
+_theWorkerPropertyDict = _WorkerPropertyDict()
+
+
+@implementer(IRenderable)
+class _SecretRenderer:
+
+    def __init__(self, secret_name):
+        self.secret_name = secret_name
+
+    @defer.inlineCallbacks
+    def getRenderingFor(self, properties):
+        secretsSrv = properties.master.namedServices.get("secrets")
+        if not secretsSrv:
+            error_message = "secrets service not started, need to configure" \
+                            " SecretManager in c['services'] to use 'secrets'" \
+                            "in Interpolate"
+            raise KeyError(error_message)
+        credsservice = properties.master.namedServices['secrets']
+        secret_detail = yield credsservice.get(self.secret_name)
+        if secret_detail is None:
+            raise KeyError("secret key {} is not found in any provider".format(self.secret_name))
+        properties.useSecret(secret_detail.value, self.secret_name)
+        return secret_detail.value
+
+
+class Secret(_SecretRenderer):
+
+    def __repr__(self):
+        return "Secret({0})".format(self.secret_name)
+
+
+class _SecretIndexer:
+
+    def __contains__(self, password):
+        return True
+
+    def __getitem__(self, password):
+        return _SecretRenderer(password)
+
+
+@implementer(IRenderable)
+class _SourceStampDict(util.ComparableMixin):
 
     compare_attrs = ('codebase',)
 
     def __init__(self, codebase):
         self.codebase = codebase
 
-    def getRenderingFor(self, build):
-        ss = build.getBuild().getSourceStamp(self.codebase)
+    def getRenderingFor(self, props):
+        ss = props.getSourceStamp(self.codebase)
         if ss:
-            return ss.asDict()
-        else:
-            return {}
+            return ss
+        return {}
 
 
-class _SlaveInfoDict(util.ComparableMixin, object):
-    implements(IRenderable)
-
-    def getRenderingFor(self, build):
-        slave = build.getBuild().slavebuilder.slave
-        return slave.slave_status.getInfoAsDict()
-
-
-class _Lazy(util.ComparableMixin, object):
-    implements(IRenderable)
+@implementer(IRenderable)
+class _Lazy(util.ComparableMixin):
 
     compare_attrs = ('value',)
 
@@ -418,17 +603,17 @@ class _Lazy(util.ComparableMixin, object):
         return '_Lazy(%r)' % self.value
 
 
-class Interpolate(util.ComparableMixin, object):
+@implementer(IRenderable)
+class Interpolate(RenderableOperatorsMixin, util.ComparableMixin):
 
     """
     This is a marker class, used fairly widely to indicate that we
     want to interpolate build properties.
     """
 
-    implements(IRenderable)
     compare_attrs = ('fmtstring', 'args', 'kwargs')
 
-    identifier_re = re.compile(r'^[\w-]*$')
+    identifier_re = re.compile(r'^[\w._-]*$')
 
     def __init__(self, fmtstring, *args, **kwargs):
         self.fmtstring = fmtstring
@@ -441,14 +626,12 @@ class Interpolate(util.ComparableMixin, object):
             self.interpolations = {}
             self._parse(fmtstring)
 
-    # TODO: add case below for when there's no args or kwargs..
     def __repr__(self):
         if self.args:
             return 'Interpolate(%r, *%r)' % (self.fmtstring, self.args)
         elif self.kwargs:
             return 'Interpolate(%r, **%r)' % (self.fmtstring, self.kwargs)
-        else:
-            return 'Interpolate(%r)' % (self.fmtstring,)
+        return 'Interpolate(%r)' % (self.fmtstring,)
 
     @staticmethod
     def _parse_prop(arg):
@@ -457,10 +640,20 @@ class Interpolate(util.ComparableMixin, object):
         except ValueError:
             prop, repl = arg, None
         if not Interpolate.identifier_re.match(prop):
-            config.error("Property name must be alphanumeric for prop Interpolation '%s'" % arg)
+            config.error(
+                "Property name must be alphanumeric for prop Interpolation '{}'".format(arg))
             prop = repl = None
+
         return _thePropertyDict, prop, repl
 
+    @staticmethod
+    def _parse_secret(arg):
+        try:
+            secret, repl = arg.split(":", 1)
+        except ValueError:
+            secret, repl = arg, None
+        return _SecretIndexer(), secret, repl
+
     @staticmethod
     def _parse_src(arg):
         # TODO: Handle changes
@@ -471,27 +664,26 @@ class Interpolate(util.ComparableMixin, object):
                 codebase, attr = arg.split(":", 1)
                 repl = None
             except ValueError:
-                config.error("Must specify both codebase and attribute for src Interpolation '%s'" % arg)
+                config.error(("Must specify both codebase and attribute for "
+                              "src Interpolation '{}'").format(arg))
                 return {}, None, None
 
         if not Interpolate.identifier_re.match(codebase):
-            config.error("Codebase must be alphanumeric for src Interpolation '%s'" % arg)
+            config.error(
+                "Codebase must be alphanumeric for src Interpolation '{}'".format(arg))
             codebase = attr = repl = None
         if not Interpolate.identifier_re.match(attr):
-            config.error("Attribute must be alphanumeric for src Interpolation '%s'" % arg)
+            config.error(
+                "Attribute must be alphanumeric for src Interpolation '{}'".format(arg))
             codebase = attr = repl = None
         return _SourceStampDict(codebase), attr, repl
 
-    @staticmethod
-    def _parse_slave(arg):
+    def _parse_worker(self, arg):
         try:
-            key, repl = arg.split(":", 1)
+            prop, repl = arg.split(":", 1)
         except ValueError:
-            key, repl = arg, None
-        if not Interpolate.identifier_re.match(key):
-            config.error("Keyword must be alphanumeric for slave-info Interpolation '%s'" % arg)
-            key = repl = None
-        return _SlaveInfoDict(), key, repl
+            prop, repl = arg, None
+        return _theWorkerPropertyDict, prop, repl
 
     def _parse_kw(self, arg):
         try:
@@ -499,7 +691,8 @@ class Interpolate(util.ComparableMixin, object):
         except ValueError:
             kw, repl = arg, None
         if not Interpolate.identifier_re.match(kw):
-            config.error("Keyword must be alphanumeric for kw Interpolation '%s'" % arg)
+            config.error(
+                "Keyword must be alphanumeric for kw Interpolation '{}'".format(arg))
             kw = repl = None
         return _Lazy(self.kwargs), kw, repl
 
@@ -507,20 +700,20 @@ class Interpolate(util.ComparableMixin, object):
         try:
             key, arg = fmt.split(":", 1)
         except ValueError:
-            config.error("invalid Interpolate substitution without selector '%s'" % fmt)
-            return
+            config.error(
+                "invalid Interpolate substitution without selector '{}'".format(fmt))
+            return None
 
         fn = getattr(self, "_parse_" + key, None)
         if not fn:
-            config.error("invalid Interpolate selector '%s'" % key)
+            config.error("invalid Interpolate selector '{}'".format(key))
             return None
-        else:
-            return fn(arg)
+        return fn(arg)
 
     @staticmethod
     def _splitBalancedParen(delim, arg):
         parenCount = 0
-        for i in range(0, len(arg)):
+        for i, val in enumerate(arg):
             if arg[i] == "(":
                 parenCount += 1
             if arg[i] == ")":
@@ -558,7 +751,8 @@ class Interpolate(util.ComparableMixin, object):
         try:
             truePart, falsePart = self._splitBalancedParen(delim, repl[1:])
         except ValueError:
-            config.error("invalid Interpolate ternary expression '%s' with delimiter '%s'" % (repl[1:], repl[0]))
+            config.error("invalid Interpolate ternary expression '{}' with delimiter '{}'".format(
+                repl[1:], repl[0]))
             return None
         return _Lookup(d, kw,
                        hasKey=Interpolate(truePart, **self.kwargs),
@@ -572,7 +766,7 @@ class Interpolate(util.ComparableMixin, object):
     def _parse(self, fmtstring):
         keys = _getInterpolationList(fmtstring)
         for key in keys:
-            if not key in self.interpolations:
+            if key not in self.interpolations:
                 d, kw, repl = self._parseSubstitution(key)
                 if repl is None:
                     repl = '-'
@@ -587,31 +781,29 @@ class Interpolate(util.ComparableMixin, object):
                     if not junk and matches:
                         self.interpolations[key] = fn(d, kw, tail)
                         break
-                if not key in self.interpolations:
-                    config.error("invalid Interpolate default type '%s'" % repl[0])
+                if key not in self.interpolations:
+                    config.error("invalid Interpolate default type '{}'".format(repl[0]))
 
-    def getRenderingFor(self, props):
-        props = props.getProperties()
+    def getRenderingFor(self, build):
+        props = build.getProperties()
         if self.args:
             d = props.render(self.args)
             d.addCallback(lambda args:
                           self.fmtstring % tuple(args))
-            return d
         else:
             d = props.render(self.interpolations)
             d.addCallback(lambda res:
                           self.fmtstring % res)
-            return d
+        return d
 
 
-class Property(util.ComparableMixin):
+@implementer(IRenderable)
+class Property(RenderableOperatorsMixin, util.ComparableMixin):
 
     """
     An instance of this class renders a property of a build.
     """
 
-    implements(IRenderable)
-
     compare_attrs = ('key', 'default', 'defaultWhenFalse')
 
     def __init__(self, key, default=None, defaultWhenFalse=True):
@@ -626,6 +818,9 @@ class Property(util.ComparableMixin):
         self.default = default
         self.defaultWhenFalse = defaultWhenFalse
 
+    def __repr__(self):
+        return "Property({0})".format(self.key)
+
     def getRenderingFor(self, props):
         if self.defaultWhenFalse:
             d = props.render(props.getProperty(self.key))
@@ -634,22 +829,20 @@ class Property(util.ComparableMixin):
             def checkDefault(rv):
                 if rv:
                     return rv
-                else:
-                    return props.render(self.default)
-            return d
-        else:
-            if props.hasProperty(self.key):
-                return props.render(props.getProperty(self.key))
-            else:
                 return props.render(self.default)
+            return d
+
+        if props.hasProperty(self.key):
+            return props.render(props.getProperty(self.key))
+        return props.render(self.default)
 
 
-class FlattenList(util.ComparableMixin):
+@implementer(IRenderable)
+class FlattenList(RenderableOperatorsMixin, util.ComparableMixin):
 
     """
     An instance of this class flattens all nested lists in a list
     """
-    implements(IRenderable)
 
     compare_attrs = ('nestedlist')
 
@@ -664,37 +857,63 @@ class FlattenList(util.ComparableMixin):
     def getRenderingFor(self, props):
         d = props.render(self.nestedlist)
 
+        @d.addCallback
         def flat(r):
             return flatten(r, self.types)
-        d.addCallback(flat)
         return d
 
+    def __add__(self, b):
+        if isinstance(b, FlattenList):
+            b = b.nestedlist
+        return FlattenList(self.nestedlist + b, self.types)
 
-class _Renderer(util.ComparableMixin, object):
-    implements(IRenderable)
 
-    compare_attrs = ('getRenderingFor',)
+@implementer(IRenderable)
+class _Renderer(util.ComparableMixin):
+
+    compare_attrs = ('fn',)
 
     def __init__(self, fn):
-        self.getRenderingFor = fn
+        self.fn = fn
+        self.args = []
+        self.kwargs = {}
+
+    def withArgs(self, *args, **kwargs):
+        new_renderer = _Renderer(self.fn)
+        new_renderer.args = self.args + list(args)
+        new_renderer.kwargs = dict(self.kwargs)
+        new_renderer.kwargs.update(kwargs)
+        return new_renderer
+
+    @defer.inlineCallbacks
+    def getRenderingFor(self, props):
+        args = yield props.render(self.args)
+        kwargs = yield props.render(self.kwargs)
+
+        # We allow the renderer fn to return a renderable for convenience
+        result = yield self.fn(props, *args, **kwargs)
+        result = yield props.render(result)
+        return result
 
     def __repr__(self):
-        return 'renderer(%r)' % (self.getRenderingFor,)
+        if self.args or self.kwargs:
+            return 'renderer(%r, args=%r, kwargs=%r)' % (self.fn, self.args,
+                                                         self.kwargs)
+        return 'renderer(%r)' % (self.fn,)
 
 
 def renderer(fn):
     return _Renderer(fn)
 
 
-class _DefaultRenderer(object):
+@implementer(IRenderable)
+class _DefaultRenderer:
 
     """
     Default IRenderable adaptor. Calls .getRenderingFor if available, otherwise
     returns argument unchanged.
     """
 
-    implements(IRenderable)
-
     def __init__(self, value):
         try:
             self.renderer = value.getRenderingFor
@@ -704,34 +923,34 @@ class _DefaultRenderer(object):
     def getRenderingFor(self, build):
         return self.renderer(build)
 
+
 registerAdapter(_DefaultRenderer, object, IRenderable)
 
 
-class _ListRenderer(object):
+@implementer(IRenderable)
+class _ListRenderer:
 
     """
     List IRenderable adaptor. Maps Build.render over the list.
     """
 
-    implements(IRenderable)
-
     def __init__(self, value):
         self.value = value
 
     def getRenderingFor(self, build):
         return defer.gatherResults([build.render(e) for e in self.value])
 
+
 registerAdapter(_ListRenderer, list, IRenderable)
 
 
-class _TupleRenderer(object):
+@implementer(IRenderable)
+class _TupleRenderer:
 
     """
     Tuple IRenderable adaptor. Maps Build.render over the tuple.
     """
 
-    implements(IRenderable)
-
     def __init__(self, value):
         self.value = value
 
@@ -740,23 +959,49 @@ class _TupleRenderer(object):
         d.addCallback(tuple)
         return d
 
+
 registerAdapter(_TupleRenderer, tuple, IRenderable)
 
 
-class _DictRenderer(object):
+@implementer(IRenderable)
+class _DictRenderer:
 
     """
-    Dict IRenderable adaptor. Maps Build.render over the keya and values in the dict.
+    Dict IRenderable adaptor. Maps Build.render over the keys and values in the dict.
     """
 
-    implements(IRenderable)
-
     def __init__(self, value):
-        self.value = _ListRenderer([_TupleRenderer((k, v)) for k, v in value.iteritems()])
+        self.value = _ListRenderer(
+            [_TupleRenderer((k, v)) for k, v in value.items()])
 
     def getRenderingFor(self, build):
         d = self.value.getRenderingFor(build)
         d.addCallback(dict)
         return d
 
+
 registerAdapter(_DictRenderer, dict, IRenderable)
+
+
+@implementer(IRenderable)
+class Transform:
+
+    """
+    A renderable that combines other renderables' results using an arbitrary function.
+    """
+
+    def __init__(self, function, *args, **kwargs):
+        if not callable(function) and not IRenderable.providedBy(function):
+            config.error(
+                "function given to Transform neither callable nor renderable")
+
+        self._function = function
+        self._args = args
+        self._kwargs = kwargs
+
+    @defer.inlineCallbacks
+    def getRenderingFor(self, iprops):
+        rfunction = yield iprops.render(self._function)
+        rargs = yield iprops.render(self._args)
+        rkwargs = yield iprops.render(self._kwargs)
+        return rfunction(*rargs, **rkwargs)
